from functools import wraps
from typing import Callable, Dict, Any, Literal, Optional, Type, Union
from pydantic import BaseModel, ValidationError, create_model
import openai
import jinja2

# Setup logging
import logging

logger = logging.getLogger("moorellm")

from moorellm.utils import _add_transitions, _create_response_model
from moorellm.models import (
    MooreRun,
    MooreState,
    DefaultResponse,
    StateMachineError,
    ImmediateStateChange,
)


class MooreFSM:
    """Moore Finite State Machine based LLM Agent class, this allows to define states and transitions and use latest
    structured response from OpenAI/Azure OpenAI API

    :param initial_state: Initial state of the FSM
    :param end_state: End state of the FSM, default is "END" (currently not used)
    :type initial_state: str
    :type end_state: str
    :return: MooreFSM object
    :rtype: :class:`moorellm.main.MooreFSM`
    """

    async def _default_get_completion(
        self,
        async_openai_instance: Union[openai.AsyncAzureOpenAI, openai.AsyncOpenAI],
        chat_history: list,
        response_model: Type[BaseModel],
        llm_model: str,
    ):
        completion = await async_openai_instance.beta.chat.completions.parse(
            model=llm_model,
            messages=chat_history,
            response_format=response_model,
        )
        
        message = completion.choices[0].message
        if not message.parsed:
            raise StateMachineError(
                f"Error in parsing the completion: {message.refusal}"
            )

        return message.parsed.model_dump()

    def __init__(self, initial_state: str, end_state: str = "END"):
        """Initialize the Moore FSM with initial state and end state"""
        self._state = initial_state
        self._next_state = None
        self._initial_state = initial_state
        self._end_state = end_state
        self._states = {}
        self._chat_history = []
        self._full_chat_history = []
        self._running_chat_history = []
        self.user_defined_context = {}
        self._is_urgent_shift = False
        
        self._get_completion = self._default_get_completion

    def override_get_completion(self, get_completion: Callable):
        """Override the get completion function with custom function,
        the result should be dictionary with response and next_state_key.

        :param get_completion: Custom function to get completion
        :type get_completion: Callable
        :return: None
        :rtype: None

        .. code-block:: python

            async def custom_get_completion(async_openai_instance, chat_history, response_model, llm_model):
                return await async_openai_instance.beta.chat.completions.parse(
                    model=llm_model,
                    messages=chat_history,
                    response_format=response_model,
                )
                        
                message = completion.choices[0].message
                if not message.parsed:
                    raise StateMachineError(
                        f"Error in parsing the completion: {message.refusal}"
                    )

                return message.parsed.model_dump()

            fsm.override_get_completion(custom_get_completion)
        """
        self._get_completion = get_completion

    def state(
        self,
        state_key: str,
        system_prompt: str,
        temperature: float = 0.5,
        transitions: Dict[str, str] = {},
        response_model: Optional[BaseModel] = None,
        pre_process_input: Optional[Callable] = None,
        pre_process_chat: Optional[Callable] = None,
        pre_process_system_prompt: Optional[Callable] = None,
    ):
        """Decorator to add a state to the FSM while defining transitions and other details.

        :param state_key: Key of this state
        :param system_prompt: System prompt for this state
        :param temperature: Temperature for the completion
        :param transitions: Transitions to other states, defined by key-value pairs of next_state_key: condition
        :param response_model: Pydantic model for response parsing, default is None (ie give string response)
        :param pre_process_input: Pre-process user input before sending to OpenAI API
        :param pre_process_chat: Pre-process chat history before sending to OpenAI API
        :param pre_process_system_prompt: Pre-process system prompt before sending to OpenAI API
        :type state_key: str
        :type system_prompt: str
        :type temperature: float
        :type transitions: Dict[str, str]
        :type response_model: BaseModel, optional
        :type pre_process_input: Callable, optional
        :type pre_process_chat: Callable, optional
        :type pre_process_system_prompt: Callable, optional
        :return: Decorator function
        :rtype: Callable

        .. code-block:: python

            @fsm.state(
                state_key="START",
                system_prompt="Hello, how can I help you?",
                temperature=0.5,
                transitions={"END": "when user has said quit or exit"},
                response_model=DefaultResponse,
            )
            async def start_state(fsm: MooreFSM, response: DefaultResponse, will_transition: bool):
                pass

        .. note:: The state function should be an async function and should have the following signature:

            .. code-block:: python

                    async def state_function(fsm: MooreFSM, response: Any, will_transition: bool):
                        pass

        .. note:: The response model should be a Pydantic model, if not defined then the response will be a string.
        """

        def decorator(func: Callable):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await func(*args, **kwargs)

            # Register the state function
            self._states[state_key] = MooreState(
                key=state_key,
                func=wrapper,
                system_prompt=system_prompt,
                temperature=temperature,
                transitions=transitions,
                response_model=response_model,
                pre_process_input=pre_process_input,
                pre_process_chat=pre_process_chat,
                pre_process_system_prompt=pre_process_system_prompt,
            )
            return wrapper

        return decorator

    async def run(
        self,
        async_openai_instance: Union[openai.AsyncAzureOpenAI, openai.AsyncOpenAI],
        user_input: str,
        model: str = "gpt-4o-2024-08-06",
        is_generator: bool = False,
        *args,
        **kwargs,
    ) -> MooreRun:
        """Run the FSM with user input and get the response from OpenAI API, only one iteration is done at a time.

        :param async_openai_instance: OpenAI/AzureOpenAI instance to use for completion
        :param user_input: User input to the FSM
        :param model: Model to use for completion, default is "gpt-4o-2024-08-06"
        :type async_openai_instance: Union[openai.AsyncAzureOpenAI, openai.AsyncOpenAI]
        :type user_input: str
        :type model: str
        :return: MooreRun object
        :rtype: :class:`moorellm.models.MooreRun`

        .. code-block:: python

            while True:
                user_input = input("You: ")
                run: MooreRun = await fsm.run(async_openai_instance, user_input)
                print(f"AI: {run.response}")
                if run.state == "END":
                    break

        .. note:: The run function should be called in a loop to keep the FSM running, it will return the details from the run.

        .. note:: Only use AsyncOpenAI or AsyncAzureOpenAI instance for async completion.
        """
        logger.debug(f"Current state: {self._state}, User input: {user_input}")

        # Get the current state
        current_state: MooreState = self._states.get(self._state, None)
        if not current_state:
            logger.error(StateMachineError(f"State {self._state} not found in states."))

        if current_state.pre_process_input:
            user_input = current_state.pre_process_input(user_input, self) or user_input
            logger.debug(f"Pre-processed user input: {user_input}")

        state_system_prompt = current_state.system_prompt

        # Pre-process system prompt with Jinja2
        template = jinja2.Template(state_system_prompt)
        state_system_prompt = template.render(self.user_defined_context)

        if current_state.pre_process_system_prompt:
            state_system_prompt = (
                current_state.pre_process_system_prompt(state_system_prompt, self)
                or state_system_prompt
            )

        processed_system_prompt = _add_transitions(state_system_prompt, current_state)
        logger.debug(f"Processed system prompt: {processed_system_prompt}")

        # Make copy of chat history (this is to prevent duplication of messages when in pattern of call then cancel)
        chat_history_copy = self._chat_history.copy()
        full_chat_history_copy = self._full_chat_history.copy()

        # Add user input to chat history
        chat_history_copy.append({"role": "user", "content": user_input})
        full_chat_history_copy.append({"role": "user", "content": user_input})

        # First create a good chat history
        system_prompt_lined = {"role": "system", "content": processed_system_prompt}
        chat_history_copy_executable = [system_prompt_lined] + chat_history_copy

        # Pre-process chat if needed
        if current_state.pre_process_chat:
            chat_history_copy_executable = current_state.pre_process_chat(
                chat_history_copy_executable, self
            )

        # Now let's try to call openai function
        logger.debug(f"Getting completion for model: {model}")

        output_response_model = _create_response_model(
            current_state.response_model, current_state.transitions, current_state.key
        )

        completion_res_in_dict = await self._get_completion(
            async_openai_instance,
            chat_history_copy_executable,
            output_response_model,
            model,
        )

        # Extract the response and next state key
        response_dict = completion_res_in_dict

        # Default to current state if no next state key
        next_state_key = response_dict.get("next_state_key", current_state.key)

        # Also get the response
        response = response_dict.get("response", None)

        # If no model was specificed then get content from response
        if not current_state.response_model:
            response = response.get("content", response)
        else:
            # Parse the response model into pydantic model
            try:
                response = current_state.response_model(**response)
            except ValidationError as e:
                raise StateMachineError(f"Error in parsing response model: {e}")

        # Check if next state key is valid
        if next_state_key not in self._states:
            logger.error(
                StateMachineError(
                    f"Next state {next_state_key} not found in states, resetting to current state."
                )
            )
            next_state_key = current_state.key

        self._next_state = next_state_key
        cached_next_state = self._next_state

        # Update the running chat history, ie this is for contexts
        self._running_chat_history = chat_history_copy
        cached_chat_history = chat_history_copy

        # Now we call function with all the details, but before that create function context params
        function_context = {
            "fsm": self,
            "response": response,
            "will_transition": self._state != self._next_state,
            **kwargs,
        }

        # Call the function
        final_response = await current_state.func(**function_context)
        if isinstance(final_response, ImmediateStateChange):
            self._state = final_response.next_state
            new_user_input = final_response.input
            logger.debug(
                f"Urgent shifting to: {self._next_state} and recreating response.."
            )
            self._is_urgent_shift = True
            moore_run_chained = await self.run(
                async_openai_instance, new_user_input, model, *args, **kwargs
            )
            self._is_urgent_shift = False
            if final_response.keep_original_response:
                final_response_str = response + final_response.keep_original_seperator + moore_run_chained.response
                moore_run_chained.response = final_response_str

            return moore_run_chained

        final_response_str = ""
        if final_response:
            # Add the response to chat history
            final_response_str = final_response
        else:
            if not current_state.response_model:
                final_response_str = response
            else:
                final_response_str = response
                raise StateMachineError(
                    "Response model is defined but no response returned from the state function."
                )

        logger.debug(f"Final response: {final_response_str}")

        # Add the response to chat history
        chat_history_copy.append({"role": "assistant", "content": final_response_str})
        full_chat_history_copy.append(
            {"role": "assistant", "content": final_response_str}
        )

        if self._running_chat_history != cached_chat_history:
            logger.debug(f"Manually set chat history: {self._running_chat_history}")
            self._chat_history = self._running_chat_history
        else:
            self._chat_history = chat_history_copy

        self._full_chat_history = full_chat_history_copy

        # Update the state
        previous_state = self._state
        self._state = self._next_state

        if self._next_state != cached_next_state:
            logger.debug(f"Manually transitioned to next state: {self._state}")
        elif self._state != previous_state:
            logger.debug(f"Transitioned to next state: {self._state}")
            self._full_chat_history.append(
                {
                    "role": "moorellm",
                    "content": f"Transitioned to next state: {self._state}",
                }
            )

        self._next_state = None

        # Return the run object
        return MooreRun(
            state=self._state,
            chat_history=chat_history_copy,
            context_data=self.user_defined_context,
            response_raw=response_dict,
            response=final_response_str,
        )

    def reset(self):
        """Reset the FSM to initial state."""
        self._state = self._initial_state
        self._next_state = None
        self._chat_history = []
        self.user_defined_context = {}
        logger.debug("FSM reset to initial state.")

    def get_next_state(self):
        """Get the next state."""
        return self._next_state

    def get_current_state(self):
        """Get the current state."""
        return self._state

    def set_next_state(self, next_state: str):
        """Set the next state."""
        self._next_state = next_state
        logger.debug(f"Manually set next state: {self._next_state}")

    def get_chat_history(self):
        """Get the chat history, not recommended to use, please use get_running_chat_history."""
        return self._chat_history

    def set_chat_history(self, chat_history: list):
        """Set the chat history, not recommended to use, please use set_running_chat_history."""
        self._chat_history = chat_history
        logger.debug(f"Chat history set: {chat_history}")

    def get_running_chat_history(self):
        """Get the running chat history."""
        return self._running_chat_history

    def set_running_chat_history(self, chat_history: list):
        """Set the running chat history."""
        self._running_chat_history = chat_history
        logger.debug(f"Running chat history set: {chat_history}")

    def get_full_chat_history(self):
        """Get the full chat history."""
        return self._full_chat_history

    def set_context_data(self, key: str, value: Any):
        """Set data into user defined context."""
        self.user_defined_context[key] = value
        logger.debug(f"User defined context set: {key}={value}")

    def set_context_data_dict(self, data: Dict[str, Any]):
        """Set data into user defined context."""
        self.user_defined_context.update(data)
        logger.debug(f"User defined context set: {data}")

    def get_context_data(self, key: str, default: Any = None):
        """Get data from user defined context."""
        return self.user_defined_context.get(key, default)

    def get_full_context_data(self):
        """Get full user defined context."""
        return self.user_defined_context

    def is_completed(self):
        """Check if the FSM is completed."""
        return self._state == self._end_state

    def is_urgent_shift(self):
        """Check if the FSM is in urgent shift."""
        return self._is_urgent_shift


from typing import Union
import openai
import jinja2
from pydantic import BaseModel
from moorellm.models import GuardrailResponse
import logging


# ------ Define Helper Functions ------ #
async def structured_call(
    chat_history: list,
    openai_instance: Union[openai.AzureOpenAI, openai.AsyncAzureOpenAI],
    response_model: type[BaseModel],
    model_name: str = "gpt-4o-2024-08-06",
    **kwargs,
):
    completion: openai.types.Completion = (
        await openai_instance.beta.chat.completions.parse(
            model=model_name,
            messages=chat_history,
            response_format=response_model,
        )
    )

    message = completion.choices[0].message
    if message.parsed:
        message: response_model = message.parsed
        return message
    else:
        print(message.refusal)
        raise Exception(
            "Failed to get structured_call, please try again.." + message.refusal
        )


# Guardrail transition function for moorellm
async def guardrail_transition(
    condition: str,
    state_from: str,
    state_to: str,
    chat_history: list[dict],
    openai_instance: Union[openai.AzureOpenAI, openai.AsyncAzureOpenAI],
):
    """Guardrail transition function for moorellm, this function will check if the transition is valid or not.

    :param condition: The condition for the transition
    :param state_from: The state from which the transition is happening
    :param state_to: The state to which the transition is happening
    :param chat_history: The chat history upto this point
    :param openai_instance: The openai instance to use
    :type condition: str
    :type state_from: str
    :type state_to: str
    :type chat_history: list[dict]
    :type openai_instance: Union[openai.AzureOpenAI, openai.AsyncAzureOpenAI]
    :return: The response from the guardrail
    :rtype: GuardrailResponse
    """
    system_prompt = f"""You are an AI states management system responsible for checking if transitions
                        from one state to another are valid, You observed that a state transition was requested
                        from state {state_from} to state {state_to}, Use user conversation upto this point to verify
                        if this transition is valid or not.

                        The condition for this transition is: {condition}"""

    chat_history_copy = chat_history.copy()
    # Iterate through the chat history and remove the system prompts
    for chat in chat_history_copy:
        if chat["role"] == "system":
            chat_history.remove(chat)
    # Add the system prompt
    chat_history.append({"role": "system", "content": system_prompt})

    response: GuardrailResponse = await structured_call(
        chat_history, openai_instance, GuardrailResponse
    )
    return response


async def llm_judge_response(
    response: str,
    good_or_bad_desc: str,
    openai_instance: Union[openai.AzureOpenAI, openai.AsyncAzureOpenAI],
    max_voting_runs: int = 1,
) -> bool:
    """Use LLM to judge if the response is good or bad as compared to the original response, using voting for n times.

    :param response: The response to judge
    :param good_or_bad_desc: The description of good or bad response
    :param openai_instance: The openai instance to use
    :param max_voting_runs: The number of times to vote (must be odd and greater than 0), defaults to 1
    :type response: str
    :type good_or_bad_desc: str
    :type openai_instance: Union[openai.AzureOpenAI, openai.AsyncAzureOpenAI]
    :type max_voting_runs: int, optional

    :return: The judged response (true if good, false if bad)
    :rtype: bool
    """
    if max_voting_runs % 2 == 0 or max_voting_runs <= 0:
        raise ValueError("max_voting_runs must be odd and greater than 0")
    # Using voting for n times and make LLM judge if our response is good or bad as compared to the original response
    good_votes = 0
    bad_votes = 0

    logger = logging.getLogger("moorellm")

    class JudgeResponse(BaseModel):
        is_response_good: bool
        thinking_steps: str

    system_prompt = f"""You are an AI Response judge system responsible for judging if the response is good or bad as compared to the how it is expected to be. \
                        Given the response and the description of good or bad response, you need to judge if the response is good or bad, Think clearly and provide the judgement."""

    user_content = (
        f"""Response: {response}, Good or Bad Description: {good_or_bad_desc}"""
    )

    for i in range(max_voting_runs):
        response: JudgeResponse = await structured_call(
            [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": user_content},
            ],
            openai_instance,
            JudgeResponse,
        )

        logger.debug(
            f"{i} - Response: {response.is_response_good}, Thinking Steps: {response.thinking_steps}"
        )
        if response.is_response_good:
            good_votes += 1
        else:
            bad_votes += 1

    return good_votes > bad_votes



import openai_responses
from pydantic import BaseModel
import pytest
import openai
from moorellm import MooreFSM, MooreState, DefaultResponse
from moorellm.models import MooreRun
from moorellm.utils import wrap_into_json_response


# Define the fixture
@pytest.fixture
def fsm():
    """Fixture for creating a MooreFSM instance."""
    return MooreFSM(initial_state="START")


@pytest.fixture
def openai_client():
    return openai.AsyncOpenAI(api_key="sk-fake123")


@pytest.fixture
def set_openai_response():
    def set_response(
        openai_mock: openai_responses.OpenAIMock, response: BaseModel, next_state: str
    ):
        openai_mock.beta.chat.completions.create.response = {
            "choices": [
                {
                    "index": 0,
                    "finish_reason": "stop",
                    "message": {
                        "content": wrap_into_json_response(response, next_state),
                        "role": "assistant",
                    },
                }
            ]
        }

    return set_response


# Test for simple fsm with state
@pytest.mark.asyncio
@openai_responses.mock()
async def test_jinja2(
    fsm: MooreFSM,
    openai_client: openai.AsyncOpenAI,
    set_openai_response,
    openai_mock: openai_responses.OpenAIMock,
):
    JINJA2_TEMPLATE = "Hello, {{ name }} is here."
    user_defined_variables = {"name": "John"}

    # Set variables
    fsm.set_context_data_dict(user_defined_variables)

    def assert_system_prompt(system_prompt: str, fsm: MooreFSM):
        assert system_prompt == "Hello, John is here."

    @fsm.state(
        state_key="START",
        system_prompt=JINJA2_TEMPLATE,
        pre_process_system_prompt=assert_system_prompt,
    )
    async def start_state(fsm: MooreFSM, response: str, will_transition: bool):
        pass

    set_openai_response(openai_mock, DefaultResponse(content=""), next_state="START")

    # Run the FSM
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "START"



import json
from typing import Callable, Dict, Any, Literal, Optional, Type, Union
from pydantic import BaseModel, ValidationError, create_model

from moorellm.models import MooreState, DefaultResponse


def _create_response_model(
    current_state_model: Union[Type[BaseModel], None],
    transitions: dict[str, str],
    default_state: str,
) -> Type[BaseModel]:
    """Create a response model based on the current state model and transitions, this will be used for structured_response openai param."""

    # Extract the transition keys as a tuple for the Literal type
    transition_keys = tuple([default_state] + list(transitions.keys()))

    next_state_key_type = Literal.__getitem__(transition_keys)

    if not current_state_model:
        current_state_model = DefaultResponse

    # Dynamically create the model with response and next_state_key fields
    return create_model(
        "EnclosedResponse",
        response=(current_state_model, ...),
        next_state_key=(next_state_key_type, ...),
    )


def _add_transitions(system_prompt: str, moore_state: MooreState) -> str:
    """Add transitions to the system prompt."""
    system_prompt += f"\n\nYou are currently in {moore_state.key} and based on user input, you can transition to the following states (with conditions defined):"
    for key, value in moore_state.transitions.items():
        system_prompt += f"\n- {key}: {value}"

    system_prompt += "\n\nIn response add the state you want to transition to.. (or leave blank to stay in the current state)"
    return system_prompt


def wrap_into_json_response(data: BaseModel, next_state: str) -> BaseModel:
    dict_res = {"response": data.model_dump(), "next_state_key": next_state}

    return json.dumps(dict_res)



import openai_responses
from pydantic import BaseModel
import pytest
import openai
from moorellm import MooreFSM, MooreState, DefaultResponse
from moorellm.models import MooreRun
from moorellm.utils import wrap_into_json_response


# Define the fixture
@pytest.fixture
def fsm():
    """Fixture for creating a MooreFSM instance."""
    return MooreFSM(initial_state="START")


@pytest.fixture
def openai_client():
    return openai.AsyncOpenAI(api_key="sk-fake123")


@pytest.fixture
def set_openai_response():
    def set_response(
        openai_mock: openai_responses.OpenAIMock, response: BaseModel, next_state: str
    ):
        openai_mock.beta.chat.completions.create.response = {
            "choices": [
                {
                    "index": 0,
                    "finish_reason": "stop",
                    "message": {
                        "content": wrap_into_json_response(response, next_state),
                        "role": "assistant",
                    },
                }
            ]
        }

    return set_response


# Test the module creation
def test_module_creation(fsm):
    """Test that the module can be created."""
    assert isinstance(fsm, MooreFSM)


# Test for simple fsm with state
@pytest.mark.asyncio
@openai_responses.mock()
async def test_simple_transitions(
    fsm: MooreFSM,
    openai_client: openai.AsyncOpenAI,
    set_openai_response,
    openai_mock: openai_responses.OpenAIMock,
):
    AI_RESPONSE_SUCCESS_TURNON = "The light is turned on."
    AI_RESPONSE_SUCCESS_TURNOFF = "The light is turned off."
    AI_RESPONSE_IDLE = "I am idle, tell me to turn on or off the light."
    AI_RESPONSE_IDLE_ON = "I am idle, tell me when to turn off the light."

    @fsm.state(
        state_key="START",
        system_prompt="You are on/off light switching AI",
        transitions={
            "STATE_ON": "If user says to turn on light",
        },
    )
    async def start_state(fsm: MooreFSM, response: str, will_transition: bool):
        if not will_transition:
            assert response == AI_RESPONSE_IDLE
        if will_transition and fsm.get_next_state() == "STATE_ON":
            assert response == AI_RESPONSE_SUCCESS_TURNON

        return response

    @fsm.state(
        state_key="STATE_ON",
        system_prompt="Turning on the light...",
        transitions={"START": "If user says to turn off light"},
    )
    async def state_on(fsm: MooreFSM, response: str, will_transition: bool):
        if will_transition and fsm.get_next_state() == "START":
            assert response == AI_RESPONSE_SUCCESS_TURNOFF
        else:
            assert response == AI_RESPONSE_IDLE_ON

    set_openai_response(
        openai_mock, DefaultResponse(content=AI_RESPONSE_IDLE), next_state="START"
    )

    # Run the FSM
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "START"
    assert run_state.response == AI_RESPONSE_IDLE

    # Now test the transitions
    set_openai_response(
        openai_mock,
        DefaultResponse(content=AI_RESPONSE_SUCCESS_TURNON),
        next_state="STATE_ON",
    )
    run_state: MooreRun = await fsm.run(openai_client, user_input="Turn on the light")
    assert run_state.state == "STATE_ON"
    assert run_state.response == AI_RESPONSE_SUCCESS_TURNON

    set_openai_response(
        openai_mock, DefaultResponse(content=AI_RESPONSE_IDLE_ON), next_state="STATE_ON"
    )
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "STATE_ON"
    assert run_state.response == AI_RESPONSE_IDLE_ON

    set_openai_response(
        openai_mock,
        DefaultResponse(content=AI_RESPONSE_SUCCESS_TURNOFF),
        next_state="START",
    )

    run_state: MooreRun = await fsm.run(openai_client, user_input="Turn off the light")
    assert run_state.state == "START"
    assert run_state.response == AI_RESPONSE_SUCCESS_TURNOFF



from typing import Any, Callable, Optional, Type
import pydantic
from pydantic import BaseModel


class MooreState(pydantic.BaseModel):
    """
    Represents a state in the Finite State Machine (FSM) for managing the AI's conversation flow.

    :param key: Unique identifier for the state.
    :param func: Callable function that defines the action to take in this state.
    :param system_prompt: The system prompt to be sent to the model.
    :param temperature: The temperature setting for the model's response generation.
    :param transitions: A dictionary mapping possible user inputs to the next state.
    :param response_model: The Pydantic model that will parse the AI's response, if provided.
    :param pre_process_input: Optional callable for pre-processing the user input before running the state function.
    :param pre_process_chat: Optional callable for pre-processing the chat input before running the state function.
    :param pre_process_system_prompt: Optional callable for pre-processing the system prompt before it is sent.

    :type key: str
    :type func: Callable
    :type system_prompt: str
    :type temperature: float
    :type transitions: dict[str, str]
    :type response_model: Type[BaseModel], optional
    :type pre_process_input: Callable, optional
    :type pre_process_chat: Callable, optional
    :type pre_process_system_prompt: Callable, optional

    .. note:: The `transitions` dictionary should map input keys to corresponding state keys for proper FSM flow.
    """

    key: str
    func: Callable
    system_prompt: str
    temperature: float
    transitions: dict[str, str]
    response_model: Optional[Type[BaseModel]]
    pre_process_input: Optional[Callable]
    pre_process_chat: Optional[Callable]
    pre_process_system_prompt: Optional[Callable]


class DefaultResponse(BaseModel):
    """
    Default response model for AI output.

    :param content: The content of the response from the AI model.

    :type content: str

    .. note:: This model can be extended or replaced with custom response models as needed.
    """

    content: str


class MooreRun(pydantic.BaseModel):
    """
    Represents the outcome of a single run/step in the FSM.

    :param state: The current state key of the FSM.
    :param chat_history: A list of dictionaries representing the history of the conversation.
    :param context_data: Contextual data relevant to the FSM run.
    :param response_raw: The raw response from the AI model.
    :param response: The processed response, potentially modeled by `response_model`.

    :type state: str
    :type chat_history: list[dict]
    :type context_data: dict[str, Any]
    :type response_raw: dict
    :type response: Any

    .. note:: The `response` attribute may be of any type, depending on the response model used.
    """

    state: str
    chat_history: list[dict]
    context_data: dict[str, Any]
    response_raw: dict
    response: Any


class StateMachineError(Exception):
    """
    Custom exception for errors within the FSM.

    .. note:: Raise this exception to indicate errors specific to FSM operations.
    """

    pass


class GuardrailResponse(BaseModel):
    message: str
    is_valid: bool


class ImmediateStateChange(BaseModel):
    """
    Change the state immediately without running the current state function.
    """

    next_state: str
    input: str = "Hey"
    keep_original_response: bool = False # If true, the original response will be preserved and prepended to the new response.
    keep_original_seperator: str = " " # Seperator to use between original and new response


import openai_responses
from pydantic import BaseModel
import pytest
import openai
from moorellm import MooreFSM, MooreState, DefaultResponse
from moorellm.models import MooreRun
from moorellm.utils import wrap_into_json_response


# Define the fixture
@pytest.fixture
def fsm():
    """Fixture for creating a MooreFSM instance."""
    return MooreFSM(initial_state="START")


@pytest.fixture
def openai_client():
    return openai.AsyncOpenAI(api_key="sk-fake123")


@pytest.fixture
def set_openai_response():
    def set_response(
        openai_mock: openai_responses.OpenAIMock, response: BaseModel, next_state: str
    ):
        openai_mock.beta.chat.completions.create.response = {
            "choices": [
                {
                    "index": 0,
                    "finish_reason": "stop",
                    "message": {
                        "content": wrap_into_json_response(response, next_state),
                        "role": "assistant",
                    },
                }
            ]
        }

    return set_response


# Test the module creation
def test_module_creation(fsm):
    """Test that the module can be created."""
    assert isinstance(fsm, MooreFSM)


# Test for simple fsm with state
@pytest.mark.asyncio
@openai_responses.mock()
async def test_single_state(
    fsm, openai_client, set_openai_response, openai_mock: openai_responses.OpenAIMock
):
    """Test that a single state can be defined."""

    COMMON_RESPONSE = "My name is Moore."

    @fsm.state(state_key="START", system_prompt="Hello, what's your name?")
    async def start_state(fsm: MooreFSM, response: str, will_transition: bool):
        assert response == COMMON_RESPONSE
        assert will_transition == False
        assert fsm.get_next_state() == "START"

    set_openai_response(
        openai_mock, DefaultResponse(content=COMMON_RESPONSE), next_state="START"
    )

    # Run the FSM
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "START"
    assert run_state.response == COMMON_RESPONSE


@pytest.mark.asyncio
@openai_responses.mock()
async def test_response_manipulation(
    fsm, openai_client, set_openai_response, openai_mock: openai_responses.OpenAIMock
):
    """Test that the response can be manipulated in a state."""

    NATURAL_RESPONSE = "My name is Moore."
    MANIPULATED_RESPONSE = "My name is Moore. What's yours?"

    @fsm.state(state_key="START", system_prompt="Hello, what's your name?")
    async def start_state(fsm: MooreFSM, response: str, will_transition: bool):
        assert response == NATURAL_RESPONSE
        assert will_transition == False
        assert fsm.get_next_state() == "START"
        return MANIPULATED_RESPONSE

    set_openai_response(
        openai_mock, DefaultResponse(content=NATURAL_RESPONSE), next_state="START"
    )

    # Run the FSM
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "START"
    assert run_state.response == MANIPULATED_RESPONSE


# Test for simple fsm with models
@pytest.mark.asyncio
@openai_responses.mock()
async def test_single_state_with_model(
    fsm, openai_client, set_openai_response, openai_mock: openai_responses.OpenAIMock
):
    """Test that response model can be generated and then given to a state."""

    class NameResponse(BaseModel):
        name: str

    EXTRACTED_NAME = "Moore"
    GREETING_PREFIX = "Hello, Welcome: "

    @fsm.state(
        state_key="START",
        system_prompt="You are an AI which extracts user's name, ask the user for their name.",
        response_model=NameResponse,
    )
    async def start_state(fsm: MooreFSM, response: NameResponse, will_transition: bool):
        assert response.name == "Moore"
        assert will_transition == False
        assert fsm.get_next_state() == "START"
        return GREETING_PREFIX + response.name

    set_openai_response(openai_mock, NameResponse(name="Moore"), next_state="START")

    # Run the FSM
    run_state: MooreRun = await fsm.run(openai_client, user_input="Hello")
    assert run_state.state == "START"
    assert run_state.response == GREETING_PREFIX + EXTRACTED_NAME



import pytest
from moorellm import MooreFSM


@pytest.fixture
def fsm():
    """Fixture for creating a MooreFSM instance."""
    return MooreFSM(initial_state="START")


def test_module_creation(fsm):
    """Test that the module can be created."""
    assert isinstance(fsm, MooreFSM)


def test_reset(fsm):
    """Test that the FSM can be reset."""
    fsm.set_next_state("NEXT")
    fsm.reset()
    assert fsm.get_current_state() == "START"


def test_get_next_state(fsm):
    """Test that the next state can be retrieved."""
    fsm.set_next_state("NEXT")
    assert fsm.get_next_state() == "NEXT"


def test_get_current_state(fsm):
    """Test that the current state can be retrieved."""
    assert fsm.get_current_state() == "START"
